#!/usr/bin/python
# -*- coding:utf8 -*-

""" IRC bot to kickban spammers on c3 channels """

import threading
import random
import time
import json
import re
import os
import signal
import sys

# Yup, I'va tweaked irclib to handle SIGHUP
sys.path.insert(0, "/home/vincent/scripts/python-myirclib")
import irclib
import ircbot

#: Bot configuration
import config
#: Module de réponse aux questions de base
import isit

# la partie qui réfère au fichier lui-même est mieux ici
# sinon on réfère la config et pas le fichier lui-même
config.thisfile = os.path.realpath(__file__)

def get_config_logfile(serveur):
    """Renvoie le nom du fichier de log en fonction du ``serveur`` et de la config."""
    serveurs = {"irc.hackint.org" : "hackint",
                "irc.crans.org" : "crans"}
    return config.logfile_template % (serveurs[serveur],)

def log(serveur, channel, auteur=None, message=None):
    """Enregistre une ligne de log."""
    if auteur == message == None:
        # alors c'est que c'est pas un channel mais juste une ligne de log
        chain = u"%s %s" % (time.strftime("%F %T"), channel)
    else:
        chain = u"%s [%s:%s] %s" % (time.strftime("%F %T"), channel, auteur, message)
    f = open(get_config_logfile(serveur), "a")
    f.write((chain + u"\n").encode("utf-8"))
    f.close()
    if config.debug_stdout:
        print chain.encode("utf-8")

def ignore_event(serv, ev):
    """Retourne ``True`` si il faut ignorer cet évènement."""
    for (blackmask, exceptlist) in config.blacklisted_masks:
        usermask = ev.source()
        blackit = bool(irclib.mask_matches(usermask, blackmask))
        exceptit = any([bool(irclib.mask_matches(usermask, exceptmask)) for exceptmask in exceptlist])
        if exceptit: # Il est exempté
            return False
        else:
            if blackit: # Il n'est pas exempté et matche la blacklist
                return True


def bot_unicode(chain):
    """Essaye de décoder ``chain`` en UTF-8.
       Lève une py:class:`errors.UnicodeBotError` en cas d'échec."""
    try:
        return chain.decode("utf8")
    except UnicodeDecodeError as exc:
        raise errors.UnicodeBotError


class Josh(ircbot.SingleServerIRCBot):
    """Classe principale : définition du bot"""
    def __init__(self, serveur, debug=False):
        temporary_pseudo = config.irc_pseudo + str(random.randrange(10000,100000))
        ircbot.SingleServerIRCBot.__init__(self, [(serveur, 6667)],
                              temporary_pseudo, "I'm a bot. Ask [20-100].", 10)
        self.debug = debug
        self.serveur = serveur
        self.overops = config.overops
        self.ops = self.overops + config.ops
        self.report_bugs_to = config.report_bugs_to
        self.chanlist = config.chanlist
        self.stay_channels = config.stay_channels
        self.quiet_channels = config.quiet_channels
        self.last_perdu = 0

    ### Utilitaires
    def _getnick(self):
        """Récuère le nick effectif du bot sur le serveur."""
        return self.serv.get_nickname()
    nick = property(_getnick)
    
    def give_me_my_pseudo(self, serv):
        """Récupère le pseudo auprès de NickServ."""
        serv.privmsg("NickServ", "RECOVER %s %s" % (config.irc_pseudo, config.irc_password))
        serv.privmsg("NickServ", "RELEASE %s %s" % (config.irc_pseudo, config.irc_password))
        time.sleep(0.3)
        serv.nick(config.irc_pseudo)
    
    def pourmoi(self, serv, message):
        """Renvoie (False, lemessage) ou (True, le message amputé de "pseudo: ")"""
        pseudo = self.nick
        pseudo = pseudo.decode("utf-8")
        size = len(pseudo)
        if message[:size] == pseudo and len(message) > size and message[size] == ":":
            return (True, message[size+1:].lstrip(" "))
        else:
            return (False, message)
    
    ### Exécution d'actions
    def quitter(self, chan, leave_message=None):
        """Quitter un channel avec un message customisable."""
        if leave_message == None:
            leave_message = random.choice(config.leave_messages)
        self.serv.part(chan, message=leave_message.encode("utf8"))
    
    def mourir(self):
        """Se déconnecter du serveur IRC avec un message customisable."""
        quit_message = random.choice(config.quit_messages)
        self.die(msg=quit_message.encode("utf8"))
    
    def execute_reload(self, auteur=None):
        """Recharge la config."""
        reload(config)
        isit.regexp_compile()
        if auteur in [None, "SIGHUP"]:
            towrite = "Config reloaded" + " (SIGHUP received)" * (auteur == "SIGHUP")
            for to in config.report_bugs_to:
                self.serv.privmsg(to, towrite)
            log(self.serveur, towrite)
            return True, None
        else:
            return True, u"Config reloaded"
    
    def crash(self, who="nobody", chan="nowhere"):
        """Fait crasher le bot."""
        where = "en privé" if chan == "priv" else "sur le chan %s" % chan
        raise errors.CrashError((u"Crash demandé par %s %s" % (who, where)).encode("utf-8"))
    
    ACTIONS = {
        "reload" : execute_reload,
        }
    
    def execute_something(self, something, params, place=None, auteur=None):
        """Exécute une action et répond son résultat à ``auteur``
           sur un chan ou en privé en fonction de ``place``"""
        action = self.ACTIONS[something]
        success, message = action(self, **params)
        if message:
            if irclib.is_channel(place):
                message = "%s: %s" % (auteur, message.encode("utf-8"))
            self.serv.privmsg(place, message)
        log(self.serveur, place, auteur, something + "%r" % params + ("[successful]" if success else "[failed]"))
    
    ### Surcharge des events du Bot
    def on_welcome(self, serv, ev):
        """À l'arrivée sur le serveur."""
        self.serv = serv # ça serv ira :)
        self.give_me_my_pseudo(serv)
        serv.privmsg("NickServ", "IDENTIFY %s" % (config.irc_password))
        log(self.serveur, "Connected")
        if self.debug:
            self.chanlist = ["#bot"]
        for c in self.chanlist:
            log(self.serveur, "JOIN %s" % (c))
            serv.join(c)

    def on_privmsg(self, serv, ev):
        """À la réception d'un message en privé."""
        if ignore_event(serv, ev):
            return
        message = ev.arguments()[0]
        auteur = irclib.nm_to_n(ev.source())
        try:
            message = bot_unicode(message)
        except errors.UnicodeBotError:
            if config.utf8_trigger:
                serv.privmsg(auteur, random.choice(config.utf8_fail_answers).encode("utf8"))
            return
        message = message.split()
        cmd = message[0].lower()
        notunderstood = False
        if cmd == u"help":
            op,overop=auteur in self.ops, auteur in self.overops
            if len(message)==1:
                helpmsg = config.helpmsg_default
                if op:
                    helpmsg += config.helpmsg_ops
                if overop:
                    helpmsg += config.helpmsg_overops
            else:
                helpmsgs = config.helpdico.get(message[1].lower(), ["Commande inconnue.", None, None])
                helpmsg = helpmsgs[0]
                if op and helpmsgs[1]:
                    if helpmsg:
                        helpmsg += "\n" + helpmsgs[1]
                    else:
                        helpmsg = helpmsgs[1]
                if overop and helpmsgs[2]:
                    if helpmsg:
                        helpmsg += "\n" + helpmsgs[2]
                    else:
                        helpmsg = helpmsgs[2]
            for ligne in helpmsg.split("\n"):
                serv.privmsg(auteur, ligne.encode("utf-8"))
        elif cmd == u"join":
            if auteur in self.ops:
                if len(message) > 1:
                    if message[1] in self.chanlist:
                        serv.privmsg(auteur, (u"Je suis déjà sur %s" % (message[1])).encode("utf-8"))
                    else:
                        serv.join(message[1])
                        self.chanlist.append(message[1])
                        serv.privmsg(auteur, "Channels : " + " ".join(self.chanlist))
                        log(self.serveur, "priv", auteur, " ".join(message))
                else:
                    serv.privmsg(auteur, "Channels : " + " ".join(self.chanlist))
            else:
                notunderstood = True
        elif cmd == u"leave":
            if auteur in self.ops and len(message) > 1:
                if message[1] in self.chanlist:
                    if not (message[1] in self.stay_channels) or auteur in self.overops:
                        self.quitter(message[1].encode("utf-8"), " ".join(message[2:]))
                        self.chanlist.remove(message[1])
                        log(self.serveur, "priv", auteur, " ".join(message) + "[successful]")
                    else:
                        serv.privmsg(auteur, "Non, je reste !")
                        log(self.serveur, "priv", auteur, " ".join(message) + "[failed]")
                else:
                    serv.privmsg(auteur, "Je ne suis pas sur %s" % (message[1]))
            else:
                notunderstood = True
        elif cmd == u"stay":
            if auteur in self.overops:
                if len(message) > 1:
                    if message[1] in self.stay_channels:
                        log(self.serveur, "priv", auteur, " ".join(message) + "[failed]")
                        serv.privmsg(auteur, "Je stay déjà sur %s." % (message[1]))
                    else:
                        log(self.serveur, "priv", auteur, " ".join(message) + "[successful]")
                        self.stay_channels.append(message[1])
                        serv.privmsg(auteur, "Stay channels : " + " ".join(self.stay_channels))
                else:
                    serv.privmsg(auteur, "Stay channels : " + " ".join(self.stay_channels))
            else:
                notunderstood = True
        elif cmd == u"nostay":
            if auteur in self.overops:
                if len(message) > 1:
                    if message[1] in self.stay_channels:
                        log(self.serveur, "priv", auteur, " ".join(message) + "[successful]")
                        self.stay_channels.remove(message[1])
                        serv.privmsg(auteur, "Stay channels : " + " ".join(self.stay_channels))
                    else:
                        log(self.serveur, "priv", auteur, " ".join(message) + "[failed]")
                        serv.privmsg(auteur, "Je ne stay pas sur %s." % (message[1]))
            else:
                notunderstood = True
        elif cmd == u"die":
            if auteur in self.overops:
                log(self.serveur, "priv", auteur, " ".join(message) + "[successful]")
                self.mourir()
            else:
                notunderstood = True
        elif cmd == u"crash":
            if auteur in self.overops:
                log(self.serveur, "priv", auteur, " ".join(message) + "[successful]")
                self.crash(auteur, "priv")
            else:
                notunderstood = True
        elif cmd == u"reload":
            if auteur in self.ops:
                self.execute_something("reload", {"auteur" : auteur}, place=auteur, auteur=auteur)
            else:
                notunderstood = True
        elif cmd == u"quiet":
            if auteur in self.ops:
                if len(message) > 1:
                    if message[1] in self.quiet_channels:
                        serv.privmsg(auteur, "Je me la ferme déjà sur %s" % (message[1]))
                        log(self.serveur, "priv", auteur, " ".join(message) + "[failed]")
                    else:
                        self.quiet_channels.append(message[1])
                        serv.privmsg(auteur, "Quiet channels : " + " ".join(self.quiet_channels))
                        log(self.serveur, "priv", auteur, " ".join(message) + "[successful]")
                else:
                    serv.privmsg(auteur, "Quiet channels : " + " ".join(self.quiet_channels))
            else:
                notunderstood = True
        elif cmd == u"noquiet":
            if auteur in self.ops:
                if len(message) > 1:
                    if message[1] in self.quiet_channels:
                        self.quiet_channels.remove(message[1])
                        serv.privmsg(auteur, "Quiet channels : " + " ".join(self.quiet_channels))
                        log(self.serveur, "priv", auteur, " ".join(message) + "[successful]")
                    else:
                        serv.privmsg(auteur, "Je ne me la ferme pas sur %s." % (message[1]))
                        log(self.serveur, "priv", auteur, " ".join(message) + "[failed]")
            else:
                notunderstood = True
        elif cmd == u"say":
            if auteur in self.overops and len(message) > 2:
                serv.privmsg(message[1].encode("utf-8"), (u" ".join(message[2:])).encode("utf-8"))
                log(self.serveur, "priv", auteur, " ".join(message))
            elif len(message) <= 2:
                serv.privmsg(auteur, "Syntaxe : SAY <channel> <message>")
            else:
                notunderstood = True
        elif cmd == u"do":
            if auteur in self.overops and len(message) > 2:
                serv.action(message[1], " ".join(message[2:]))
                log(self.serveur, "priv", auteur, " ".join(message))
            elif len(message) <= 2:
                serv.privmsg(auteur, "Syntaxe : DO <channel> <action>")
            else:
                notunderstood = True
        elif cmd == u"kick":
            if auteur in self.overops and len(message) > 2:
                serv.kick(message[1].encode("utf-8"), message[2].encode("utf-8"), " ".join(message[3:]).encode("utf-8"))
                log(self.serveur, "priv", auteur, " ".join(message))
            elif len(message) <= 2:
                serv.privmsg(auteur, "Syntaxe : KICK <channel> <pseudo> [<raison>]")
            else:
                notunderstood = True
        elif cmd == u"ops":
            if auteur in self.overops:
                serv.privmsg(auteur, " ".join(self.ops))
            else:
                notunderstood = True
        elif cmd == u"overops":
            if auteur in self.overops:
                serv.privmsg(auteur, " ".join(self.overops))
            else:
                notunderstood = True
        if notunderstood:
            serv.privmsg(auteur, "I don't understand what you're saying. Try HELP…")
    
    def on_pubmsg(self, serv, ev):
        """À la réception d'un message sur un channel."""
        if ignore_event(serv, ev):
            return
        auteur = irclib.nm_to_n(ev.source())
        canal = ev.target()
        message = ev.arguments()[0]
        try:
            message = bot_unicode(message)
        except errors.UnicodeBotError:
            if config.utf8_trigger and not canal in self.quiet_channels:
                serv.privmsg(canal, (u"%s: %s"% ( auteur, random.choice(config.utf8_fail_answers))).encode("utf8"))
            return
        pour_moi, message = self.pourmoi(serv, message)
        if pour_moi and message.split()!=[]:
            cmd = message.split()[0].lower()
            try:
                args = " ".join(message.split()[1:])
            except:
                args = ""
            if cmd in [u"meurs", u"die", u"crève"]:
                if auteur in self.overops:
                    log(self.serveur, canal, auteur, message + "[successful]")
                    self.mourir()
                else:
                    serv.privmsg(canal,(u"%s: %s"%(auteur, random.choice(config.quit_fail_messages))).encode("utf8"))
                    log(self.serveur, canal, auteur, message + "[failed]")
            elif cmd == u"reload":
                if auteur in self.ops:
                    self.execute_something("reload", {"auteur" : auteur}, place=canal, auteur=auteur)
            elif cmd == u"crash":
                if auteur in self.overops:
                    self.crash(auteur, canal)
            elif cmd in [u"part", u"leave", u"dégage", u"va-t-en"]:
                if auteur in self.ops and (not (canal in self.stay_channels)
                                           or auteur in self.overops):
                    self.quitter(canal)
                    log(self.serveur, canal, auteur, message + "[successful]")
                    if canal in self.chanlist:
                        self.chanlist.remove(canal)
                else:
                    serv.privmsg(canal,(u"%s: %s" % (auteur, random.choice(config.leave_fail_messages))).encode("utf8"))
                    log(self.serveur, canal, auteur, message + "[failed]")
            
            elif cmd in [u"deviens", u"pseudo", u"become", u"nick"]:
                if auteur in self.ops:
                    become = args
                    serv.nick(become)
                    log(self.serveur, canal, auteur, message + "[successful]")
    
            elif cmd in [u"ping"] and not canal in self.quiet_channels:
                serv.privmsg(canal, "%s: pong" % (auteur))

            if isit.is_tag(message) and not canal in self.quiet_channels:
                if auteur in self.ops:
                    action = random.choice(config.tag_actions)
                    serv.action(canal, action.encode("utf8"))
                    self.quiet_channels.append(canal)
                else:
                    answer = random.choice(config.tag_answers)
                    for ligne in answer.split("\n"):
                        serv.privmsg(canal, "%s: %s" % (auteur, ligne.encode("utf8")))
        if isit.is_allah(message):
            serv.kick(canal, auteur, u"Please go spread your proselytism somewhere else.")

    def on_action(self, serv, ev):
        """À la réception d'une action."""
        if ignore_event(serv, ev):
            return
        action = ev.arguments()[0]
        auteur = irclib.nm_to_n(ev.source())
        channel = ev.target()
        try:
            action = bot_unicode(action)
        except errors.UnicodeBotError:
            if config.utf8_trigger and not channel in self.quiet_channels:
                serv.privmsg(channel, (u"%s: %s"%(auteur,random.choice(config.utf8_fail_answers))).encode("utf8"))
            return
        mypseudo = self.nick
        
    def on_kick(self, serv, ev):
        """À la réception d'une action."""
        auteur = irclib.nm_to_n(ev.source())
        channel = ev.target()
        victime = ev.arguments()[0]
        raison = ev.arguments()[1]
        if victime == self.nick:
            log(self.serveur, ("%s kické de %s par %s (raison : %s)" % (victime, channel, auteur, raison)).decode("utf-8"))
            time.sleep(10)
            serv.join(channel)
#            l1, l2 = config.kick_answers, config.kick_actions
#            n1, n2 = len(l1), len(l2)
#            i = random.randrange(n1 + n2)
#            if i >= n1:
#                serv.action(channel, l2[i-n1].format(auteur).encode("utf8"))
#            else:
#                serv.privmsg(channel, l1[i].format(auteur).encode("utf8"))
    
    ### .fork trick
    def start_as_daemon(self, outfile):
        sys.stderr = Logger(outfile)
        self.start()
    

class Logger(object):
    """Pour écrire ailleurs que sur stdout"""
    def __init__(self, filename="josh.full.log"):
        self.filename = filename

    def write(self, message):
        f = open(self.filename, "a")
        f.write(message)
        f.close()

def main():
    """Exécution principal : lecture des paramètres et lancement du bot."""
    if len(sys.argv) == 1:
        print "Usage : josh.py <serveur> [--debug] [--no-output] [--daemon [--pidfile]] [--outfile]"
        print "        --outfile sans --no-output ni --daemon n'a aucun effet"
        exit(1)
    serveur = sys.argv[1]
    if "--daemon" in sys.argv:
        thisfile = os.path.realpath(__file__)
        thisdirectory = thisfile.rsplit("/", 1)[0]
        os.chdir(thisdirectory)
        daemon = True
    else:
        daemon = False
    if "debug" in sys.argv or "--debug" in sys.argv:
        debug = True
    else:
        debug = False
    if "--quiet" in sys.argv:
        config.debug_stdout = False
    serveurs = {
        "crans" : "irc.crans.org",
        "irc.crans.org" : "irc.crans.org",
        "hackint" : "irc.hackint.org",
        "irc.hackint.org" : "irc.hackint.org",
        }
    if "--no-output" in sys.argv or "--daemon" in sys.argv:
        outfile = "/var/log/bots/josh.full.log"
        for arg in sys.argv:
            arg = arg.split("=")
            if arg[0].strip('-') in ["out", "outfile", "logfile"]:
                outfile = arg[1]
        sys.stdout = Logger(outfile)
    try:
        serveur = serveurs[serveur]
    except KeyError:
        print "Server Unknown : %s" % (serveur)
        exit(404)
    josh = Josh(serveur,debug)
    # Si on reçoit un SIGHUP, on reload la config
    def sighup_handler(signum, frame):
        josh.execute_reload(auteur="SIGHUP")
    signal.signal(signal.SIGHUP, sighup_handler)
    # Daemonization
    if daemon:
        child_pid = os.fork()
        if child_pid == 0:
            os.setsid()
            josh.start_as_daemon(outfile)
        else:
            # on enregistre le pid du bot
            pidfile = "/var/run/bots/josh.pid"
            for arg in sys.argv:
                arg = arg.split("=")
                if arg[0].strip('-') in ["pidfile"]:
                    pidfile = arg[1]
            f = open(pidfile, "w")
            f.write("%s\n" % child_pid)
            f.close()
    else:
        josh.start()

if __name__ == "__main__":
    main()
